C======================================================================
C  
C    File: /home/ama-iris/g/fortranm/Berna-Mesh/NX/mesh_lib.F
C
C    Last Modified By: Rajit Manohar (rajit@vlsi.cs.caltech.edu)
C    Modification Date: Thu Jul 21 03:12:07 PDT 1994
C    Modification: Converted to NX.
C
C    Original Author: Berna Massingill (berna@vlsi.cs.caltech.edu)
C    Date: Thu Jul 21 03:11:27 PDT 1994
C    Info: MESH TEMPLATE --- 2-D Version.
C
C=======================================================================

C=======================================================================
C=======================================================================
C
C	main programs 
C
C=======================================================================
C=======================================================================

C=======================================================================
C
C	main program -- define and connect ports, start processes
C
C=======================================================================

	program MAIN

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	logical error

C=======safety checks

        if (numnodes() .lt. NPROCS) then
           print *, 'SORRY: PROGRAM ONLY WORKS FOR ', NPROCS,
     -          ' PROCESSES.'
           stop
        endif

C	check for invalid combinations of dimensions and
C	process-grid dimensions

	error = .FALSE.

	if (NX - (XPROCS-1)*NXlsize .le. 0) then
		print*, 'invalid combination of NX, XPROCS'
		error = .TRUE.
	endif
	if (NY - (YPROCS-1)*NYlsize .le. 0) then
		print*, 'invalid combination of NY, YPROCS'
		error = .TRUE.
	endif
	if (error) STOP

C	check that the designated i/o process is the one we 
C	assume in the processes block below

	if (.not. lpio(ipxnum(1),ipynum(1))) then
		print*, 'internal error:  wrong procnum for I/O process'
		STOP
	endif

	if (mynode() .lt. NPROCS) then
	call pxpypz (mynode()+1, ipx, ipy)
	if (lpio(ipx,ipy)) then
           call mesh_io_proc(ipx,ipy)
	else
           call mesh_grid_proc(ipx,ipy)
        endif
	endif

	end

C=======================================================================
C
C	main program for i/o process -- copy arguments into 
C		common block
C
C=======================================================================

        subroutine mesh_io_proc(a_iprocx, a_iprocy)

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	integer a_iprocx, a_iprocy

	iprocx = a_iprocx
	iprocy = a_iprocy

	iproc_idx = iprocnum(iprocx, iprocy)

	l_ioproc = .true.

	NXlocal = NXlsize
	IF (iprocx .EQ. XPROCS) NXlocal = NX-(XPROCS-1)*NXlsize
	NYlocal = NYlsize
	IF (iprocy .EQ. YPROCS) NYlocal = NY-(YPROCS-1)*NYlsize

	call procmain

	return
	end

C=======================================================================
C
C	main program for grid process -- copy arguments into 
C		common block
C
C=======================================================================

        subroutine mesh_grid_proc(a_iprocx, a_iprocy)

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	integer a_iprocx, a_iprocy

	iprocx = a_iprocx
	iprocy = a_iprocy

	iproc_idx = iprocnum(iprocx, iprocy)

	l_ioproc = .false.

	NXlocal = NXlsize
	IF (iprocx .EQ. XPROCS) NXlocal = NX-(XPROCS-1)*NXlsize
	NYlocal = NYlsize
	IF (iprocy .EQ. YPROCS) NYlocal = NY-(YPROCS-1)*NYlsize

	call procmain

	return
	end

C=======================================================================
C=======================================================================
C
C	subroutine to update ghost boundaries
C
C=======================================================================
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine mesh_update_bdry(grid_array)
C
C	precondition:
C		D is a distributed array of size NX by NY.
C		in each process, grid_array = local section of D.
C		all processes have called
C			mesh_update_bdry(grid_array).
C	postcondition:
C		in each process, ghost boundaries of grid_array
C			are "correct", including corners but excluding
C			external boundaries.
C		(SAY THIS MORE CLEARLY?)
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	real grid_array(IXLO:IXHI, IYLO:IYHI)

C--	REVISE THIS??
C	here we take the most simple and straightforward approach --
C	separate buffers for each communication axis and a lot of
C	copying -- no doubt there are more efficient approaches.

	real xbuffer(NGHOST, IYLO:IYHI)
	real ybuffer(IXLO:IXHI, NGHOST)

	integer XBUFFSIZE, YBUFFSIZE
	parameter (XBUFFSIZE = (NGHOST) * (IYHI-IYLO+1) * REAL_SIZE)
	parameter (YBUFFSIZE = (IXHI-IXLO+1) * (NGHOST) * REAL_SIZE)

C	update x-low boundaries
	if (iprocx .lt. XPROCS) then
		do mm = 1, NGHOST
		do jj = 1, NYlocal
			xbuffer(mm,jj) =
     -				grid_array(NXlsize+1-mm,jj)
		enddo
		enddo
                idest = iprocnum (iprocx+1, iprocy)-1
                call csend (MSG_XCH_XHI_TAG, xbuffer, XBUFFSIZE,
     -               idest, 0)
	endif
	if (iprocx .gt. 1) then
                call crecv (MSG_XCH_XHI_TAG, xbuffer, XBUFFSIZE)
		do mm = 1, NGHOST
		do jj = 1, NYlocal
			grid_array(1-mm,jj) = xbuffer(mm,jj)
		enddo
		enddo
	endif

C	update x-high boundaries
	if (iprocx .gt. 1) then
		do mm = 1, NGHOST
		do jj = 1, NYlocal
			xbuffer(mm,jj) = grid_array(mm,jj)
		enddo
		enddo
                idest = iprocnum (iprocx-1, iprocy)-1
                call csend (MSG_XCH_XLOW_TAG, xbuffer, XBUFFSIZE,
     -               idest, 0)
	endif
	if (iprocx .lt. XPROCS) then
                call crecv (MSG_XCH_XLOW_TAG, xbuffer, XBUFFSIZE)
		do mm = 1, NGHOST
		do jj = 1, NYlocal
			grid_array(NXlsize+mm,jj) = xbuffer(mm,jj)
		enddo
		enddo
	endif

C	update y-low boundaries, including x bdry data just updated
	if (iprocy .lt. YPROCS) then
		do mm = 1, NGHOST
		do ii = 1-NGHOST, NXlocal+NGHOST
			ybuffer(ii,mm) = grid_array(ii,NYlsize+1-mm)
		enddo
		enddo
                idest = iprocnum (iprocx, iprocy+1)-1
                call csend (MSG_XCH_YHI_TAG, ybuffer, YBUFFSIZE,
     -               idest, 0)
	endif
	if (iprocy .gt. 1) then
                call crecv (MSG_XCH_YHI_TAG, ybuffer, YBUFFSIZE)
		do mm = 1, NGHOST
		do ii = 1-NGHOST, NXlocal+NGHOST
			grid_array(ii,1-mm) = ybuffer(ii,mm)
		enddo
		enddo
	endif

C	update y-high boundaries, including x bdry data just updated
	if (iprocy .gt. 1) then
		do mm = 1, NGHOST
		do ii = 1-NGHOST, NXlocal+NGHOST
			ybuffer(ii,mm) = grid_array(ii,mm)
		enddo
		enddo
                idest = iprocnum (iprocx, iprocy-1)-1
                call csend (MSG_XCH_YLOW_TAG, ybuffer, YBUFFSIZE,
     -               idest, 0)
	endif
	if (iprocy .lt. YPROCS) then
                call crecv (MSG_XCH_YLOW_TAG, ybuffer, YBUFFSIZE)
		do mm = 1, NGHOST
		do ii = 1-NGHOST, NXlocal+NGHOST
			grid_array(ii,NYlsize+mm) = ybuffer(ii,mm)
		enddo
		enddo
	endif

	return
	end


C=======================================================================
C=======================================================================
C
C     MODIFIED BOUNDARY EXCHANGE: DOES EXCHANGE OF GLOBAL BOUNDARIES.
C
C=======================================================================
C=======================================================================
C#######################################################################
C
C    Modification By: Rajit Manohar (rajit@vlsi.cs.caltech.edu)
C    On: Wed Jul 20 23:55:26 PDT 1994
C
C
C       Different from usual mesh update boundary, since it exchanges
C       even the global boundaries. Needed for periodic boundary 
C       conditions.
C#######################################################################
C-----------------------------------------------------------------------
C
	subroutine mesh_update_bdry_all(grid_array)
C
C	precondition:
C		D is a distributed array of size NX by NY.
C		in each process, grid_array = local section of D.
C		all processes have called
C			mesh_update_bdry(grid_array).
C	postcondition:
C		in each process, ghost boundaries of grid_array
C			are "correct", including corners and including
C			external boundaries.
C		(SAY THIS MORE CLEARLY?)
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	real grid_array(IXLO:IXHI, IYLO:IYHI)

C--	REVISE THIS??
C	here we take the most simple and straightforward approach --
C	separate buffers for each communication axis and a lot of
C	copying -- no doubt there are more efficient approaches.

	real xbuffer(NGHOST, IYLO:IYHI)
	real ybuffer(IXLO:IXHI, NGHOST)

	integer XBUFFSIZE, YBUFFSIZE
	parameter (XBUFFSIZE = (NGHOST) * (IYHI-IYLO+1) * REAL_SIZE)
	parameter (YBUFFSIZE = (IXHI-IXLO+1) * (NGHOST) * REAL_SIZE)

C	update x-low boundaries
	   do mm = 1, NGHOST
	      do jj = 1, NYlocal
		 xbuffer(mm,jj) =
     -                grid_array(NXlsize+1-mm,jj)
              enddo
           enddo
           idest = iprocnum (mod(iprocx,XPROCS)+1, iprocy)-1
           call csend (MSG_XCH_XHI_TAG, xbuffer, XBUFFSIZE,idest, 0)
           call crecv (MSG_XCH_XHI_TAG, xbuffer, XBUFFSIZE)
           do mm = 1, NGHOST
              do jj = 1, NYlocal
                 grid_array(1-mm,jj) = xbuffer(mm,jj)
              enddo
           enddo

C	update x-high boundaries
           do mm = 1, NGHOST
              do jj = 1, NYlocal
                 xbuffer(mm,jj) = grid_array(mm,jj)
              enddo
           enddo
           idest = iprocnum (mod(iprocx+XPROCS-2, XPROCS)+1, iprocy)-1
           call csend (MSG_XCH_XLOW_TAG, xbuffer, XBUFFSIZE,idest, 0)
           call crecv (MSG_XCH_XLOW_TAG, xbuffer, XBUFFSIZE)
           do mm = 1, NGHOST
              do jj = 1, NYlocal
                 grid_array(NXlsize+mm,jj) = xbuffer(mm,jj)
              enddo
           enddo

C	update y-low boundaries, including x bdry data just updated
           do mm = 1, NGHOST
              do ii = 1-NGHOST, NXlocal+NGHOST
                 ybuffer(ii,mm) = grid_array(ii,NYlsize+1-mm)
              enddo
           enddo
           idest = iprocnum (iprocx,mod(iprocy,YPROCS)+1)-1
           call csend (MSG_XCH_YHI_TAG, ybuffer, YBUFFSIZE,idest, 0)
           call crecv (MSG_XCH_YHI_TAG, ybuffer, YBUFFSIZE)
           do mm = 1, NGHOST
              do ii = 1-NGHOST, NXlocal+NGHOST
                 grid_array(ii,1-mm) = ybuffer(ii,mm)
              enddo
           enddo

C	update y-high boundaries, including x bdry data just updated
           do mm = 1, NGHOST
              do ii = 1-NGHOST, NXlocal+NGHOST
                 ybuffer(ii,mm) = grid_array(ii,mm)
              enddo
           enddo
           idest = iprocnum (iprocx, mod(iprocy+YPROCS-2,YPROCS)+1)-1
           call csend (MSG_XCH_YLOW_TAG, ybuffer, YBUFFSIZE,idest, 0)
           call crecv (MSG_XCH_YLOW_TAG, ybuffer, YBUFFSIZE)
           do mm = 1, NGHOST
              do ii = 1-NGHOST, NXlocal+NGHOST
                 grid_array(ii,NYlsize+mm) = ybuffer(ii,mm)
              enddo
           enddo

	return
	end

C=======================================================================
C=======================================================================
C
C	global-reduction subroutines
C
C=======================================================================
C=======================================================================

C=======================================================================
C
C	subroutine to get integer maximum
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine mesh_merge_int_max(isize, int_in, int_out)
C
C	precondition:
C		int_in and int_out are integer arrays of size isize,
C			with int_in[P] denoting the value of int_in
C			in process P.
C		all processes have called
C			mesh_merge_int_max(isize, int_in, int_out).
C	postcondition:
C		in all processes, int_out is the
C			(elementwise) maximum over int_in[P], where
C			P ranges over all processes.
C		in the i/o process, int_in has unspeicified value
C			(used as work space).
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	integer isize
	integer int_in(isize), int_out(isize)

	if (.not. l_ioproc) then
                call csend (MSG_GTOHINT_TAG, int_in,
     -			isize*INTEGER_SIZE, 0, 0)
                call crecv (MSG_HTOGINT_TAG, int_out,
     -			isize*INTEGER_SIZE)
	else
C		compute maxima into int_out in i/o process
		do mm = 1, isize
			int_out(mm) = int_in(mm)
		enddo

C		loop over processes in the same order as the sequential
C		version
		do iproc = 1, NPROCS
		if (iproc .ne. iproc_idx) then
                        call crecv (MSG_GTOHINT_TAG, int_in,
     -				isize*INTEGER_SIZE)
			do mm = 1, isize
				int_out(mm) = max(int_out(mm),int_in(mm))
			enddo
		endif
		enddo

C		send back to other processes
		do ii = 1, XPROCS
		do jj = 1, YPROCS
			itemp = iprocnum(ii, jj)
			if (itemp .ne. iproc_idx) then
				idest = itemp - 1
                        	call csend (MSG_HTOGINT_TAG, int_out,
     -					 isize*INTEGER_SIZE, idest, 0)
			endif
		enddo
		enddo
	endif

	return
	end

C=======================================================================
C
C	subroutine to get element with greatest absolute value (real)
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine mesh_merge_real_maxabs(isize, real_in, real_out)
C
C	precondition:
C		real_in and real_out are real arrays of size isize,
C			with real_in[P] denoting the value of real_in
C			in process P.
C		all processes have called
C			mesh_merge_real_maxabs(isize, real_in, real_out).
C	postcondition:
C		in all processes, real_out is the
C			(elementwise) maximum absolute value over
C			real_in[P], where P ranges over all processes.
C		in the i/o process, real_in has unspeicified value
C			(used as work space).
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	integer isize
	real real_in(isize), real_out(isize)
        integer iarray(16)

	if (.not. l_ioproc) then
                call csend (MSG_GTOHFLT_TAG, real_in, isize*REAL_SIZE,
     -			0, 0)
                call crecv (MSG_HTOGFLT_TAG, real_out, isize*REAL_SIZE)
	else
C		compute maxima into real_out in i/o process
		do mm = 1, isize
			real_out(mm) = real_in(mm)
		enddo
	
C		loop over processes in the same order as the sequential
C		version
		do iproc = 1, NPROCS
		if (iproc .ne. iproc_idx) then
                        call crecv (MSG_GTOHFLT_TAG, real_in,
     -				isize*REAL_SIZE)
			do mm = 1, isize
				if (abs(real_in(mm)) .gt. abs(real_out(mm)))
     -					real_out(mm) = real_in(mm)
			enddo
		endif
		enddo

C		send back to other processes
		do ii = 1, XPROCS
		do jj = 1, YPROCS
			itemp = iprocnum(ii, jj)
			if (itemp .ne. iproc_idx) then
				idest = itemp - 1
                        	call csend (MSG_HTOGFLT_TAG, real_out,
     -					isize*REAL_SIZE, idest, 0)
			endif
		enddo
		enddo
	endif

	return
	end

C=======================================================================
C
C	subroutine to get element with greatest value (real)
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine mesh_merge_real_max(isize, real_in, real_out)
C
C	precondition:
C		real_in and real_out are real arrays of size isize,
C			with real_in[P] denoting the value of real_in
C			in process P.
C		all processes have called
C			mesh_merge_real_max(isize, real_in, real_out).
C	postcondition:
C		in all processes, real_out is the
C			(elementwise) maximum over real_in[P],
C			where P ranges over all processes.
C		in the i/o process, int_in has unspeicified value
C			(used as work space).
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	integer isize
	real real_in(isize), real_out(isize)
        integer iarray(16)

	if (.not. l_ioproc) then
                call csend (MSG_GTOHFLT_TAG, real_in, isize*REAL_SIZE,
     -				0, 0)
                call crecv (MSG_HTOGFLT_TAG, real_out, isize*REAL_SIZE)
	else
C		compute maxima into real_out in i/o process
		do mm = 1, isize
			real_out(mm) = real_in(mm)
		enddo
	
C		loop over processes in the same order as the sequential
C		version
		do iproc = 1, NPROCS
		if (iproc .ne. iproc_idx) then
                        call crecv (MSG_GTOHFLT_TAG, real_in,
     -				isize*REAL_SIZE)
			do mm = 1, isize
				if (real_in(mm) .gt. real_out(mm)) then
     					real_out(mm) = real_in(mm)
				endif
			enddo
		endif
		enddo

C		send back to other processes
		do ii = 1, XPROCS
		do jj = 1, YPROCS
			itemp = iprocnum(ii, jj)
			if (itemp .ne. iproc_idx) then
				idest = itemp - 1
                        	call csend (MSG_HTOGFLT_TAG, real_out,
     -					isize*REAL_SIZE, idest, 0)
			endif
		enddo
		enddo
	endif

	return
	end

C=======================================================================
C
C	subroutine to get real sum
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine mesh_merge_real_sum(isize, real_in, real_out)
C
C	precondition:
C		real_in and real_out are real arrays of size isize,
C			with real_in[P] denoting the value of real_in
C			in process P.
C		all processes have called
C			mesh_merge_real_sum(isize, real_in, real_out).
C	postcondition:
C		in all processes, real_out is the
C			(elementwise) sum[P], where P ranges over all
C			processes.
C		in the i/o process, int_in has unspeicified value
C			(used as work space).
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	integer isize
	real real_in(isize), real_out(isize)
        integer iarray(16)

	if (.not. l_ioproc) then
                call csend (MSG_GTOHFLT_TAG, real_in, isize*REAL_SIZE,
     -			0, 0)
                call crecv (MSG_HTOGFLT_TAG, real_out, isize*REAL_SIZE)
	else
C		compute sums into real_out in i/o process
		do mm = 1, isize
			real_out(mm) = real_in(mm)
		enddo
	
C		loop over processes in the same order as the sequential
C		version
		do iproc = 1, NPROCS
		if (iproc .ne. iproc_idx) then
                        call crecv (MSG_GTOHFLT_TAG, real_in,
     -				isize*REAL_SIZE)
			do mm = 1, isize
				real_out(mm) = real_out(mm) + real_in(mm)
			enddo
		endif
		enddo

C		send back to other processes
		do ii = 1, XPROCS
		do jj = 1, YPROCS
			itemp = iprocnum(ii, jj)
			if (itemp .ne. iproc_idx) then
				idest = itemp - 1
                        	call csend (MSG_HTOGFLT_TAG, real_out,
     -					isize*REAL_SIZE, idest, 0)
			endif
		enddo
		enddo
	endif

	return
	end

C=======================================================================
C
C	subroutine to synchronize all processes
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine mesh_synch
C
C	precondition:
C		all processes have called mesh_synch.
C	postcondition:
C		all processes have called mesh_synch.
C		(THAT IS -- this is a barrier synchronization.)
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	integer isynch(1)

	isynch(1) = 0
	call mesh_merge_int_max(1, isynch, idummy)
	return
	end

C=======================================================================
C=======================================================================
C
C	subroutines to broadcast data
C
C=======================================================================
C=======================================================================

C=======================================================================
C
C	subroutine to broadcast integers
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine mesh_bcast_int(isize, int_array)
C
C	precondition:
C		int_array is an integer array of size isize, with
C			int_array[P] denoting the value of int_array
C			in process P and int_array[0] denoting
C			the value of int_array in the i/o process.
C		all processes have called
C			mesh_bcast_int(isize, int_array).
C	postcondition:
C		in all processes, int_array has the
C			same value as int_array[0].
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	integer isize
	integer int_array(isize)

	if (l_ioproc) then
		do ii = 1, XPROCS
		do jj = 1, YPROCS
			itemp = iprocnum(ii, jj)
			if (itemp .ne. iproc_idx) then
				idest = itemp
				call csend (MSG_HTOGINT_TAG, int_array,
     -					isize*INTEGER_SIZE, idest, 0)
			endif
		enddo
		enddo
	else
                call crecv (MSG_HTOGINT_TAG, int_array,
     -			isize*INTEGER_SIZE)
	endif

	return
	end

C-----------------------------------------------------------------------
C
	subroutine mesh_bcast_real(isize, real_array)
C
C	precondition:
C		real_array is an real array of size isize, with
C			real_array[P] denoting the value of real_array
C			in process P and real_array[0] denoting
C			the value of real_array in the i/o process.
C		all processes have called
C			mesh_bcast_real(isize, real_array).
C	postcondition:
C		in all processes, real_array has the
C			same value as real_array[0].
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	integer isize
	real real_array(isize)

	if (l_ioproc) then
		do ii = 1, XPROCS
		do jj = 1, YPROCS
			itemp = iprocnum(ii, jj)
			if (itemp .ne. iproc_idx) then
				idest = itemp - 1
				call csend (MSG_HTOGFLT_TAG, real_array,
     -					isize*REAL_SIZE, idest, 0)
			endif
		enddo
		enddo
	else
                call crecv (MSG_HTOGFLT_TAG, real_array,
     -			isize*REAL_SIZE)
	endif

	return
	end


C=======================================================================
C=======================================================================
C
C	utility functions and subroutines
C
C=======================================================================
C=======================================================================

C=======================================================================
C
C	routines to map from process grid coordinates
C		to process number and vice versa
C
C=======================================================================

C-----------------------------------------------------------------------
C
	integer function iprocnum(ipx, ipy)
C
C	precondition:
C		ipx, ipy are coordinates of process in process
C			grid.
C	postcondition:
C		return value is a single integer denoting process
C			-- unique for each process.
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'

	iprocnum = (ipy-1)*XPROCS + ipx
	return
	end

C-----------------------------------------------------------------------
C
	subroutine pxpypz(iproc, ipx, ipy)
C
C	precondition:
C		iproc is a single integer denoting process
C			-- unique for each process.
C	postcondition:
C		ipx, ipy are coordinates of process in process.
C			grid.
C		(i.e., this subroutine is the inverse of function
C			iprocnum above.)
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'

C	compute ipx - 1
	itemp = iproc - 1
	ipxm1 = mod(itemp, XPROCS)
	ipx = ipxm1 + 1
C	compute ipy - 1
	itemp = (iproc - ipx) / XPROCS
	ipym1 = mod(itemp, YPROCS)
	ipy = ipym1 + 1
	return
	end

C-----------------------------------------------------------------------
C
	integer function ipxnum(iproc)
C
C	precondition:
C		iproc is a single integer denoting process
C			-- unique for each process.
C	postcondition:
C		return value is a single integer denoting iproc's
C			x coordinate in the process grid.
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'

	call pxpypz(iproc, ipxtmp, ipytmp)
	ipxnum = ipxtmp
	return
	end

C-----------------------------------------------------------------------
C
	integer function ipynum(iproc)
C
C	precondition:
C		iproc is a single integer denoting process
C			-- unique for each process.
C	postcondition:
C		return value is a single integer denoting iproc's
C			y coordinate in the process grid.
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'

	call pxpypz(iproc, ipxtmp, ipytmp)
	ipynum = ipytmp
	return
	end

C-----------------------------------------------------------------------
C
	logical function lpio(ipx, ipy)
C
C	precondition:
C		ipx, ipy are coordinates of process in process
C			grid.
C	postcondition:
C		return value is true iff (ipx, ipy) is the
C			designated I/O process.
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'

	lpio = ipx .eq. 1 .and. ipy .eq. 1
	return
	end

C-----------------------------------------------------------------------
C
	integer function ipio()
C
C	precondition:
C		true.
C	postcondition:
C		return value is process index of designated I/O process
C			(consistent with function lpio above).
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'

	ipio = iprocnum(1,1)
	return
	end

C=======================================================================
C
C	subroutines to map from global to local indices
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine ilocal(ii, ipx, iiloc)
C
C	precondition:
C		ii is a global X index.
C	postcondition:
C		ipx, iiloc are the corresponding X index into the
C			process grid and local X index.
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	ipx = (ii-1) / NXlsize + 1
	iiloc = mod((ii-1), NXlsize) + 1
	return
	end

C-----------------------------------------------------------------------

C-----------------------------------------------------------------------
C
	subroutine jlocal(jj, ipy, jjloc)
C
C	precondition:
C		jj is a global Y index.
C	postcondition:
C		ipy, jjloc are the corresponding Y index into the
C			process grid and local Y index.
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	ipy = (jj-1) / NYlsize + 1
	jjloc = mod((jj-1), NYlsize) + 1
	return
	end

C=======================================================================
C
C	subroutines to map from local to global indices
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine iglobal(ipx, iiloc, ii)
C
C	precondition:
C		ipx, iiloc are an X index into the process grid and
C			a local X index.
C	postcondition:
C		ii is the corresponding global X index.
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	ii = (ipx-1)*NXlsize + iiloc
	return
	end

C-----------------------------------------------------------------------
C
	subroutine jglobal(ipy, jjloc, jj)
C
C	precondition:
C		ipy, jjloc are an Y index into the process grid and
C			a local Y index.
C	postcondition:
C		jj is the corresponding global Y index.
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	jj = (ipy-1)*NYlsize + jjloc
	return
	end

C=======================================================================
C
C	subroutines to determine intersection of a range of global
C	indices with the local section
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine xintersect(iglob1, iglob2, iloc1, iloc2, lempty)
C
C	precondition:
C		iglob1, iglob2 specify a range of global X indices.
C	postcondition:
C		lempty is .true. if the intersection of this range
C			with the local section is empty, .false.
C			otherwise.
C		iloc1, iloc2 specify the (local indices of the)
C			intersection of iglob1:iglob2
C			with the local section (if lempty is .false.).
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	logical lempty

C	transform global indices into local indices
	call ilocal(iglob1,ipx1,iloc1)
	call ilocal(iglob2,ipx2,iloc2)
C	is the intersection empty?
	if (ipx1 .gt. iprocx .or. ipx2 .lt. iprocx) then
		lempty = .true.
	else
		lempty = .false.
		IF (ipx1 .LT. iprocx) iloc1=1
		IF (ipx2 .GT. iprocx) iloc2=NXlocal
	endif
	return
	end

C-----------------------------------------------------------------------
C
	subroutine yintersect(jglob1, jglob2, jloc1, jloc2, lempty)
C
C	precondition:
C		jglob1, jglob2 specify a range of global X indices.
C	postcondition:
C		lempty is .true. if the intersection of this range
C			with the local section is empty, .false.
C			otherwise.
C		jloc1, jloc2 specify the (local indices of the)
C			intersection of jglob1:jglob2
C			with the local section (if lempty is .false.).
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'

	logical lempty

C	transform global indices into local indices
	call jlocal(jglob1,ipy1,jloc1)
	call jlocal(jglob2,ipy2,jloc2)
C	is the intersection empty?
	if (ipy1 .gt. iprocy .or. ipy2 .lt. iprocy) then
		lempty = .true.
	else
		lempty = .false.
		IF (ipy1 .LT. iprocy) jloc1=1
		IF (ipy2 .GT. iprocy) jloc2=NYlocal
	endif
	return
	end

C=======================================================================
C
C	subroutine to generate filename
C
C=======================================================================

C-----------------------------------------------------------------------
C
	subroutine fname(ipx, ipy, inname, outname)
C
C	precondition:
C		ipx, ipy are coordinates of process in process
C			grid -- assumed to
C			be no more than 2 digits each.
C		inname is a character string -- base for file name.
C	postcondition:
C		outname is inname followed by:
C			process grid coordinates (ipx, ipy).
C
C-----------------------------------------------------------------------

	include 'mesh_uparms.h'
	include 'mesh_parms.h'

	character*(*) inname
	character*(*) outname
	character*8 cbuff
        integer ilen

        ilen = len(inname)
        do while (inname(ilen:ilen) .eq. ' ')
            ilen = ilen - 1
        end do

	if (XPROCS .le. 9 .and. YPROCS .le. 9) then
		write (cbuff, '(3i1.1)') ipx,ipy
	else
		write (cbuff, '(3i2.2)') ipx,ipy
	endif
	outname = inname(1:ilen) // cbuff

	return
	end

C=======================================================================
C=======================================================================
C
C               Compressed Parallel I/O Routines.
C
C=======================================================================
C=======================================================================

        subroutine mesh_copenwrite (filename, ifd)
C
C       IN: filename
C             string of characters. Name of file to open for writing.
C       If the file does not exist, it is created. Any existing file
C       with the same name is deleted.
C	The length of filename MUST BE LESS THAN 128
C
C       OUT: ifd
C             Integer variable which can be used to refer to the file in
C       subsequent parallel I/O subroutine calls.
C
C       NOTE: 
C             THIS FUNCTION CAN ONLY BE CALLED FROM A GRID PROCESS.
C

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	
	character*(*) filename
	character*128 parfile
	integer ifd
	integer itype(4)

	call fname (iprocx, iprocy, filename,  parfile)
	call C_PIOOPENWRITE (ifd, parfile, 128)
	if (ifd .eq. -1) then
	   print *, 'MESH_COPENWRITE: Process: ', iprocx, iprocy, 
     -          ' Could not open file ', filename
           stop
        endif
	itype(1) = NX
	itype(2) = NY
	itype(3) = XPROCS
	itype(4) = YPROCS
	call C_PIOWRITEINT (ifd, itype, 4)
        if (ifd .eq. -1) then
	   print *, 'MESH_COPENWRITE: Process: ', iprocx, iprocy, 
     -          ' Could not write to file ', filename
           stop
        endif
	end
      
C=======================================================================

        subroutine mesh_copenread (filename, ifd)
C
C       IN: filename
C             string of characters. Name of file to open for reading.
C	The length of filename MUST BE LESS THAN 128
C
C       OUT: ifd
C             Integer variable which can be used to refer to the file in
C       subsequent parallel I/O subroutine calls.
C
C       NOTE: 
C             THIS FUNCTION CAN ONLY BE CALLED FROM A GRID PROCESS.
C

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	
	character*(*) filename
	character*128 parfile
	integer ifd
	integer itype(4)

	call fname (iprocx, iprocy, filename,  parfile)
	call C_PIOOPENREAD (ifd, parfile, 128)
	if (ifd .eq. -1) then
	   print *, 'MESH_COPENREAD: Process: ', iprocx, iprocy, 
     -          ' Could not open file ', filename
           stop
        endif
	call C_PIOREADINT (ifd, itype, 4)
        if (ifd .eq. -1) then
	   print *, 'MESH_COPENREAD: Process: ', iprocx, iprocy, 
     -          ' Could not read from file ', filename
           stop
        endif
        if ((itype(1) .ne. NX) .or. (itype(2) .ne. NY) .or.
     -       (itype(3) .ne. XPROCS) .or. (itype(4) .ne. YPROCS)) then
           print *, 'MESH_COPENREAD: Incompatible input data'
           print *, 'MESH_COPENREAD: Data was created for parameters:'
           print *, 'MESH_COPENREAD: NX=', itype(1), ' NY=', itype(2),
     -          ' XPROCS=', XPROCS, ' YPROCS=', YPROCS
           stop
        endif
	end

C=======================================================================
      
      subroutine mesh_cwrite (ifd, array)

C     IN: ifd
C         Integer. Should be a file descriptor generated by 
C     mesh_copenwrite.
C
C     IN: array
C         A REAL array, which is of size IXLO:IXHI, IYLO:IYHI. Essentially,
C     a local section.
C
C     OUT: ifd
C        If ifd is changed, it is assigned a value -1. This value
C     indicates an error status.
C
C     NOTE: 
C         This writes out the array to the data file specified by ifd.
C
C

      include 'mesh_uparms.h'
      include 'mesh_parms.h'
      include 'mesh_common.h'

      integer ifd, i
      real array(IXLO:IXHI, IYLO:IYHI)

      do i = 1, NYlocal
         call C_PIOWRITEREAL(ifd, array(1,i), NXlocal)
         if (ifd .eq. -1) then
            print *, 'MESH_CWRITE: Write to file failed.'
            return
         endif
      end do

      end

C=======================================================================
      
      subroutine mesh_cread (ifd, array)

C     IN: ifd
C         Integer. Should be a file descriptor generated by 
C     mesh_copenread.
C
C     OUT: array
C         A REAL array, which is of size IXLO:IXHI, IYLO:IYHI. Essentially,
C     a local section. The data from the file is read into this array.
C     Note that the ghost boundaries are _NOT_ updated.
C
C     OUT: ifd
C        If ifd is changed, it is assigned a value -1. This value
C     indicates an error status.
C
C     NOTE: 
C        This reads in an array from the file into "array".
C
      include 'mesh_uparms.h'
      include 'mesh_parms.h'
      include 'mesh_common.h'

      integer ifd, i
      real array(IXLO:IXHI, IYLO:IYHI)

      do i = 1, NYlocal
         call C_PIOREADREAL(ifd, array(1,i), NXlocal)
         if (ifd .eq. -1) then
            print *, 'MESH_CREAD: Read from file failed.'
            return
         endif
      end do

      end

C=======================================================================
      
      subroutine mesh_cclose (ifd)

C     IN: ifd
C         Integer. Should be a file descriptor generated by 
C     mesh_copenwrite or mesh_copenread.
C
C     OUT: ifd
C         ifd is changed to -1. 
C
C     NOTE: 
C         This writes out the array to the data file specified by ifd.
C
C

      include 'mesh_uparms.h'
      include 'mesh_parms.h'
      include 'mesh_common.h'

      integer ifd

      call C_PIOCLOSE(ifd)
      if (ifd .eq. -1) then
         print *, 'MESH_CCLOSE: Error closing file.'
         return
      endif
      ifd = -1
      
      end

C=======================================================================
C=======================================================================
C
C       Parallel I/O routines: These write data in an uncompressed format
C
C=======================================================================
C=======================================================================

        subroutine mesh_openwrite (filename, ifd)
C
C       IN: filename
C             string of characters. Name of file to open for writing.
C       If the file does not exist, it is created. Any existing file
C       with the same name is deleted.
C	The length of filename MUST BE LESS THAN 128
C
C       OUT: ifd
C             Integer variable which can be used to refer to the file in
C       subsequent parallel I/O subroutine calls.
C
C       NOTE: 
C             THIS FUNCTION CAN ONLY BE CALLED FROM A GRID PROCESS.
C

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	
	character*(*) filename
	character*128 parfile
	integer ifd
	integer itype(4)

	call fname (iprocx, iprocy, filename,  parfile)
	call PIOOPENWRITE (ifd, parfile, 128)
	if (ifd .eq. -1) then
	   print *, 'MESH_OPENWRITE: Process: ', iprocx, iprocy, 
     -          ' Could not open file ', filename
           stop
        endif
	itype(1) = NX
	itype(2) = NY
	itype(3) = XPROCS
	itype(4) = YPROCS
	call PIOWRITEINT (ifd, itype, 4)
        if (ifd .eq. -1) then
	   print *, 'MESH_OPENWRITE: Process: ', iprocx, iprocy, 
     -          ' Could not write to file ', filename
           stop
        endif
	end
      
C=======================================================================

        subroutine mesh_openread (filename, ifd)
C
C       IN: filename
C             string of characters. Name of file to open for reading.
C	The length of filename MUST BE LESS THAN 128
C
C       OUT: ifd
C             Integer variable which can be used to refer to the file in
C       subsequent parallel I/O subroutine calls.
C
C       NOTE: 
C             THIS FUNCTION CAN ONLY BE CALLED FROM A GRID PROCESS.
C

	include 'mesh_uparms.h'
	include 'mesh_parms.h'
	include 'mesh_common.h'
	
	character*(*) filename
	character*128 parfile
	integer ifd
	integer itype(4)

	call fname (iprocx, iprocy, filename,  parfile)
	call PIOOPENREAD (ifd, parfile, 128)
	if (ifd .eq. -1) then
	   print *, 'MESH_OPENREAD: Process: ', iprocx, iprocy, 
     -          ' Could not open file ', filename
           stop
        endif
	call PIOREADINT (ifd, itype, 4)
        if (ifd .eq. -1) then
	   print *, 'MESH_OPENREAD: Process: ', iprocx, iprocy, 
     -          ' Could not read from file ', filename
           stop
        endif
        if ((itype(1) .ne. NX) .or. (itype(2) .ne. NY) .or.
     -       (itype(3) .ne. XPROCS) .or. (itype(4) .ne. YPROCS)) then
           print *, 'MESH_OPENREAD: Incompatible input data'
           print *, 'MESH_OPENREAD: Data was created for parameters:'
           print *, 'MESH_OPENREAD: NX=', itype(1), ' NY=', itype(2),
     -          ' XPROCS=', XPROCS, ' YPROCS=', YPROCS
           stop
        endif
	end

C=======================================================================
      
      subroutine mesh_write (ifd, array)

C     IN: ifd
C         Integer. Should be a file descriptor generated by 
C     mesh_copenwrite.
C
C     IN: array
C         A REAL array, which is of size IXLO:IXHI, IYLO:IYHI. Essentially,
C     a local section.
C
C     OUT: ifd
C        If ifd is changed, it is assigned a value -1. This value
C     indicates an error status.
C
C     NOTE: 
C         This writes out the array to the data file specified by ifd.
C
C

      include 'mesh_uparms.h'
      include 'mesh_parms.h'
      include 'mesh_common.h'

      integer ifd, i
      real array(IXLO:IXHI, IYLO:IYHI)

      do i = 1, NYlocal
         call PIOWRITEREAL(ifd, array(1,i), NXlocal)
         if (ifd .eq. -1) then
            print *, 'MESH_WRITE: Write to file failed.'
            return
         endif
      end do

      end

C=======================================================================
      
      subroutine mesh_read (ifd, array)

C     IN: ifd
C         Integer. Should be a file descriptor generated by 
C     mesh_copenread.
C
C     OUT: array
C         A REAL array, which is of size IXLO:IXHI, IYLO:IYHI. Essentially,
C     a local section. The data from the file is read into this array.
C     Note that the ghost boundaries are _NOT_ updated.
C
C     OUT: ifd
C        If ifd is changed, it is assigned a value -1. This value
C     indicates an error status.
C
C     NOTE: 
C        This reads in an array from the file into "array".
C
      include 'mesh_uparms.h'
      include 'mesh_parms.h'
      include 'mesh_common.h'

      integer ifd, i
      real array(IXLO:IXHI, IYLO:IYHI)

      do i = 1, NYlocal
         call PIOREADREAL(ifd, array(1,i), NXlocal)
         if (ifd .eq. -1) then
            print *, 'MESH_READ: Read from file failed.'
            return
         endif
      end do

      end

C=======================================================================
      
      subroutine mesh_close (ifd)

C     IN: ifd
C         Integer. Should be a file descriptor generated by 
C     mesh_copenwrite or mesh_copenread.
C
C     OUT: ifd
C         ifd is changed to -1. 
C
C     NOTE: 
C         This writes out the array to the data file specified by ifd.
C
C

      include 'mesh_uparms.h'
      include 'mesh_parms.h'
      include 'mesh_common.h'

      integer ifd

      call PIOCLOSE(ifd)
      if (ifd .eq. -1) then
         print *, 'MESH_CLOSE: Error closing file.'
         return
      endif
      ifd = -1
      
      end
